import fnmatch

from typing import List


def is_in_blacklist(uri: str, blacklist: List[str]) -> bool:
    """
    Return True if the URI is blacklisted.

      >>> is_in_blacklist('/some-real-uri/', ['/exact-uri/'])
      False
      >>> is_in_blacklist('/exact-uri/', ['/exact-uri/'])
      True
      >>> is_in_blacklist('/job/42/retry', ['/job/*'])
      True

    """
    for pattern in blacklist:
        if fnmatch.fnmatch(uri, pattern):
            return True
    return False


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import logging
from typing import List, Optional
from datetime import datetime, timezone
from timeit import default_timer as timer
from flask.signals import request_started
from flask import Flask, request

from .core import CoreAPM
from .utils import is_in_blacklist

logger = logging.getLogger('howfast_apm')


class HowFastFlaskMiddleware(CoreAPM):
    """
    Flask middleware to measure how much time is spent per endpoint.

    This implementation is purposedly naive and potentially slow, but its goal is to validate the
    PoC. It should be replaced/improved in the future, based on the results of the PoC.
    """

    # TODO: this architecture is very ugly - there must be a better solution
    # Temporary variable holding the endpoint name between the start and the end of the request
    current_endpoint: Optional[str]

    def __init__(
            self,
            # The Flask application to analyze
            app: Flask,
            # The HowFast app ID to use
            app_id: str = None,
            # Endpoints not to monitor
            endpoints_blacklist: List[str] = None,
            # Other configuration parameters passed to the CoreAPM constructor
            **kwargs,
    ):
        super().__init__(**kwargs)

        self.app = app
        self.wsgi_app = app.wsgi_app
        # Overwrite the WSGI application
        app.wsgi_app = self

        if endpoints_blacklist:
            self.endpoints_blacklist = endpoints_blacklist
        else:
            self.endpoints_blacklist = []

        # Setup the queue and the background thread
        self.setup(app_id)

        request_started.connect(self._request_started)

    def __call__(self, environ, start_response):
        if not self.app_id:
            # HF APM not configured, return early to save some time
            return self.wsgi_app(environ, start_response)

        uri = environ.get('PATH_INFO')

        if is_in_blacklist(uri, self.endpoints_blacklist):
            # Endpoint blacklist, return now
            return self.wsgi_app(environ, start_response)

        method = environ.get('REQUEST_METHOD')

        time_request_started = datetime.now(timezone.utc)
        start = timer()
        return_value = self.wsgi_app(environ, start_response)
        end = timer()
        elapsed = end - start

        self.save_point(
            time_request_started=time_request_started,
            time_elapsed=elapsed,
            method=method,
            uri=uri,
            endpoint=self.current_endpoint,
        )
        self.current_endpoint = None
        # TODO: remove this once overhead has been measured in production
        logger.info("overhead when saving the point: %.3fms", (timer() - end) * 1000)
        return return_value

    def _request_started(self, sender, **kwargs):
        with sender.app_context():
            self.current_endpoint = request.endpoint



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['howfast_apm']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.22,<3.0']

extras_require = \
{'flask': ['flask>=0.8', 'blinker>=1.1']}

setup_kwargs = {
    'name': 'howfast-apm',
    'version': '0.4.0',
    'description': 'Lightweight Application Performance Monitoring middleware that measures and reports performance data to HowFast APM.',
    'long_description': "HowFast APM for Python servers\n==============================\n\nLight instrumentation of your Python server for reporting performance data to `HowFast APM <https://www.howfast.tech/>`_.\n\n.. image:: https://github.com/HowFast/apm-python/blob/master/screenshot.png\n    :align: center\n    :alt: Screenshot from HowFast APM\n\nInstall\n-------\n\nTo install / update the module:\n\n.. code:: bash\n\n    pip install howfast-apm[flask]\n\nUsage\n-------\n\nOnly the Flask middleware is currently available.\n\n.. code:: python\n\n    from howfast_apm import HowFastFlaskMiddleware\n\n    # Create your Flask app\n    app = Flask(__name__, ...)\n\n    # Instanciate all your other middlewares first\n\n    # Setup the APM middleware last, so that it can track the time spent inside other middlewares\n    HowFastFlaskMiddleware(app, app_id=HOWFAST_APM_DSN)\n\nConfiguration\n-------------\n\nYou can configure the APM through environment variables. If they are defined, those variables will\nbe used. Parameters passed to the ``HowFastFlaskMiddleware`` constructor take precedence over environment\nvariables.\n\nOnly one variable is available for now:\n\n* ``HOWFAST_APM_DSN``: The DSN (application identifier) that you can find on your APM dashboard. Can also be passed to the constructor as ``app_id``.\n\nIf the environment variable is defined you can then use:\n\n.. code:: python\n\n    # Install the middleware\n    HowFastFlaskMiddleware(app)\n\nYou can also choose to exclude some URLs from reporting:\n\n.. code:: python\n\n    # Do not report performance data for some URLs\n    HowFastFlaskMiddleware(\n        app,\n        endpoints_blacklist=[\n            '/some/internal/url/',\n            # You can also use patterns accepted by Python's `fnmatch.fnmatch`, shell-like:\n            '/admin/*',\n            '/jobs/*/results',\n            '/endpoint/?',  # will blacklist /endpoint and /endpoint/\n        ],\n    )\n",
    'author': 'Mickaël Bergem',
    'author_email': 'mickael@howfast.tech',
    'url': 'https://github.com/HowFast/apm-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)
